
//		DateToString.cpp -- converts a date to a string, in a variety of formats


#ifndef DATE_TO_STRING_VER
#define DATE_TO_STRING_VER			1.0.0.0


#define MIN_YEAR					1601
#define MAX_YEAR					9999
#define JULIAN_DOS_EPOCH			138427
#define JULIAN_UNIX_EPOCH			134775

#define DTS_LOCAL_NAME_LEN			40


const unsigned int MonthLength[13][2]	= { {NULL, NULL}, {31, 31}, {28, 29}, {31, 31}, {30, 30}, {31, 31}, {30, 30}, {31, 31}, {31, 31}, {30, 30}, {31, 31}, {30, 30}, {31, 31} };
const unsigned int Kalends[13][2]		= { {NULL, NULL}, {0, 0}, {31, 31}, {59, 60}, {90, 91}, {120, 121}, {151, 152}, {181, 182}, {212, 213}, {243, 244}, {273, 274}, {304, 305}, {334, 335} };

const wchar_t DowNameLong[7][10]		= { L"Sunday", L"Monday", L"Tuesday", L"Wednesday", L"Thursday", L"Friday", L"Saturday" };
const wchar_t MonthName[13][10]			= { L"", L"January", L"February", L"March", L"April", L"May", L"June", L"July", L"August", L"September", L"October", L"November", L"December" };

wchar_t LocalDowNameLong[7][DTS_LOCAL_NAME_LEN]		= { L"Sunday", L"Monday", L"Tuesday", L"Wednesday", L"Thursday", L"Friday", L"Saturday" };
wchar_t LocalMonthNameLong[13][DTS_LOCAL_NAME_LEN]	= { L"", L"January", L"February", L"March", L"April", L"May", L"June", L"July", L"August", L"September", L"October", L"November", L"December" };



int Leapness( int year )
{
	if ( ( year % 400 ) == 0 ) return 1;
	if ( ( year % 100 ) == 0 ) return 0;
	if ( ( year % 4 ) == 0 ) return 1;
	return 0;
}



long int Julian( unsigned int year, unsigned int month, unsigned int day )
{
	//  A "julian" in this program is a day number, with julian 1
	//  being January 1, 1601.  It does not correspond to the
	//  traditional, big-J Julian day number.

	if ( ( year < MIN_YEAR ) || ( year > MAX_YEAR ) )
		return -1;

	if ( ( month < 1 ) || ( month > 12 ) )
		return -1;

	BOOL leap = Leapness( year );

	if ( ( day < 1 ) || ( day > MonthLength[month][leap] ) )
		return -1;

	long int j = ( ( year - MIN_YEAR ) * 365 );
	j += ( ( year - MIN_YEAR ) / 4 );
	j -= ( ( year - MIN_YEAR ) / 100 );
	j += ( ( year - MIN_YEAR ) / 400 );
	j += Kalends[month][leap];
	j += day;

	return j;
}



int ISOWeekOffset( unsigned int year )
{
	//  Returns a value indicating where ISO week 01 begins for
	//  the given year, relative to January 1.  Possible return
	//  values range from -3 (three days prior, i.e. December 29)
	//  through +3 (three days later, i.e. January 4).  A return
	//  value of 0 would mean that January 1st is a Monday.

	if ( ( year < MIN_YEAR ) || ( year > MAX_YEAR ) ) return -999;

	int w = ( Julian( year, 1, 4 ) % 7 );
	int offset = 4 - ( ( w==0 ) ? 7 : w );

	return offset;
}



unsigned int ISOWeeksIn( unsigned int year )
{
	//  Years starting on a Thursday, and leap years starting on
	//  a Wednesday, have 53 ISO weeks.  All others have 52.

	BOOL leap = Leapness( year );
	int w = Julian( year, 1, 1 ) % 7;

	if ( ( w == 4 ) || ( ( w == 3 ) && leap ) )
		return 53;

	return 52;
}



int MakeISOWeekDate( unsigned int Year, unsigned int Month, unsigned int Day, unsigned int *ISOYear, unsigned int *ISOWeek, unsigned int *ISOWeekDay )
{
	if ( !ISOYear || !ISOWeek || !ISOWeekDay )
		return -1;

	*ISOYear = Year;
	int ord = Kalends[Month][Leapness( Year )] + Day - 1;
	int offset = ISOWeekOffset( *ISOYear );
	*ISOWeek = ( ( ord - offset ) / 7 ) + 1;
	*ISOWeekDay = ( (ord - offset + 7) % 7 ) + 1;

	if ( *ISOWeek > ISOWeeksIn( *ISOYear ) ) {
		*ISOYear += 1;
		*ISOWeek = 1;
	}
	else if ( ord < offset ) {
		*ISOYear -= 1;
		*ISOWeek = ISOWeeksIn( *ISOYear );
	}

	return 0;
}



int LocalShortDateFmt()
{
	wchar_t shortbuf[SHORT_BUF_LEN]		= L"";


	if ( GetLocaleInfo( LOCALE_USER_DEFAULT, LOCALE_IDATE, shortbuf, SHORT_BUF_LEN ) ) {
		long int i = -1;
		i = wcstol( shortbuf, NULL, 10 );
		if ( ( i >= 0 ) && ( i <= 2 ) )
			return i + 1;
	}

	return 1;
}



int LocalLongDateFmt()
{
	wchar_t shortbuf[SHORT_BUF_LEN]		= L"";


	if ( GetLocaleInfo( LOCALE_USER_DEFAULT, LOCALE_ILDATE, shortbuf, SHORT_BUF_LEN ) ) {
		long int i = -1;
		i = wcstol( shortbuf, NULL, 10 );
		if ( ( i >= 0 ) && ( i <= 2 ) )
			return i + 1;
	}

	return 1;
}



wchar_t LocalDateSepChar()
{
	wchar_t shortbuf[SHORT_BUF_LEN]		= L"";


	if ( GetLocaleInfo( LOCALE_USER_DEFAULT, LOCALE_SDATE, shortbuf, SHORT_BUF_LEN ) )
		if ( iswpunct( shortbuf[0] ) && ( shortbuf[1] == '\0' ) )
			return shortbuf[0];

	return '/';
}



int GetLocalMonthAndDowNames()
{
	BOOL Fail		= FALSE;
	int rv			= 1;


	for ( int i = 0; i < 12; i++ ) {
		LocalMonthNameLong[i + 1][0] = '\0';
		GetLocaleInfo( LOCALE_USER_DEFAULT, LOCALE_SMONTHNAME1 + i, LocalMonthNameLong[i + 1], DTS_LOCAL_NAME_LEN );
		if ( LocalMonthNameLong[i + 1][0] == '\0' )
			Fail = TRUE, rv = 0;
	}

	if ( Fail )
		for ( int i = 0; i < 13; i++ )
			wcscpy_s( LocalMonthNameLong[i], DTS_LOCAL_NAME_LEN, MonthName[i] );

	Fail = FALSE;

	for ( int i = 0; i < 7; i++ ) {
		LocalDowNameLong[i][0] = '\0';
		GetLocaleInfo( LOCALE_USER_DEFAULT, LOCALE_SDAYNAME1 + ( ( i + 6 ) % 7 ), LocalDowNameLong[i], DTS_LOCAL_NAME_LEN );
		if ( LocalDowNameLong[i][0] == '\0' )
			Fail = TRUE, rv = 0;
	}

	if ( Fail )
		for ( int i = 0; i < 7; i++ )
			wcscpy_s( LocalDowNameLong[i], DTS_LOCAL_NAME_LEN, DowNameLong[i] );

	return rv;
}



int MakeSysTime( unsigned int Year, unsigned int Month, unsigned int Day, SYSTEMTIME *outTime )
{
	if ( !outTime )
		return -1;


	memset( outTime, 0x00, sizeof( SYSTEMTIME ) );

	outTime->wYear		= Year;
	outTime->wMonth		= Month;
	outTime->wDay		= Day;
	outTime->wDayOfWeek	= Julian( Year, Month, Day ) % 7;
	return 0;
}



int DateToString( LPTSTR outString, unsigned int Year, unsigned int Month, unsigned int Day, int fmt )
{
	if ( !outString )
		return -1;

	if ( ( Year < MIN_YEAR ) || ( Year > MAX_YEAR ) || ( Month < 1 ) || ( Month > 12 ) || ( Day < 1 ) || ( Day > 31 ) )
		return -1;


	wchar_t shortbuf[SHORT_BUF_LEN]		= L"";

	BOOL leap				= Leapness( Year );
	wchar_t DateSep			= LocalDateSepChar();
	BOOL FourDigitYears		= ( fmt & 0x0100 ) > 0;
	SYSTEMTIME SysTime;
	long int j				= 0;
	long int tj				= 0;


	fmt = fmt & 0x00ff;

	if ( !fmt )
		fmt = LocalShortDateFmt();
	

	switch ( fmt ) {

	case 1 :

		if ( FourDigitYears )
			wsprintf( outString, L"%02u%c%02u%c%04u", Month, DateSep, Day, DateSep, Year );
		else
			wsprintf( outString, L"%02u%c%02u%c%02u", Month, DateSep, Day, DateSep, Year % 100 );

		break;

	case 2 :

		if ( FourDigitYears )
			wsprintf( outString, L"%02u%c%02u%c%04u", Day, DateSep, Month, DateSep, Year );
		else
			wsprintf( outString, L"%02u%c%02u%c%02u", Day, DateSep, Month, DateSep, Year % 100 );

		break;

	case 3 :

		if ( FourDigitYears )
			wsprintf( outString, L"%04u%c%02u%c%02u", Year, DateSep, Month, DateSep, Day );
		else
			wsprintf( outString, L"%02u%c%02u%c%02u", Year % 100, DateSep, Month, DateSep, Day );

		break;

	case 5 :
		{
			unsigned int isoyear = 0, isoweek = 0, isoweekday = 0;
			MakeISOWeekDate( Year, Month, Day, &isoyear, &isoweek, &isoweekday );
			wsprintf( outString, L"%04u-W%02u-%u", isoyear, isoweek, isoweekday );
		}

		break;

	case 6 :

		wsprintf( outString, L"%04u-%03u", Year, Kalends[Month][leap] + Day );
		break;

	case 16 :

		j = Julian( Year, Month, Day ) - JULIAN_DOS_EPOCH;

		if ( j < 0 )
			wsprintf( outString, L"%d", j );
		else
			wsprintf( outString, L"+%d", j );

		break;

	case 17 :

		if ( LocalLongDateFmt() == 2 )
			wsprintf( outString, L"%s, %u %s %u", DowNameLong[Julian( Year, Month, Day ) % 7], Day, MonthName[Month], Year );
		else
			wsprintf( outString, L"%s, %s %u, %u", DowNameLong[Julian( Year, Month, Day ) % 7], MonthName[Month], Day, Year );

		break;

	case 18 :

		GetLocalTime( &SysTime );
		j	= Julian( Year, Month, Day );
		tj	= Julian( SysTime.wYear, SysTime.wMonth, SysTime.wDay );

		if ( j >= tj )
			wsprintf( outString, L"T+%d", j - tj );
		else
			wsprintf( outString, L"T%d", j - tj );

		break;

	case 19 :

		wsprintf( outString, L"A+%d", Julian( Year, Month, Day ) );
		break;

	case 20 :

		wsprintf( outString, L"%04d%02d%02d", Year, Month, Day );
		break;

	case 21 :

		j = Julian( Year, Month, Day ) - JULIAN_UNIX_EPOCH;

		if ( j < 0 )
			wsprintf( outString, L"U%d", j );
		else
			wsprintf( outString, L"U+%d", j );

		break;

	case 32 :

		MakeSysTime( Year, Month, Day, &SysTime );
		GetDateFormat( LOCALE_USER_DEFAULT, DATE_LONGDATE, &SysTime, NULL, outString, SHORT_BUF_LEN );
		break;

	case 33 :

		MakeSysTime( Year, Month, Day, &SysTime );
		GetDateFormat( LOCALE_USER_DEFAULT, DATE_SHORTDATE, &SysTime, NULL, outString, SHORT_BUF_LEN );
		break;

	case 34 :

		MakeSysTime( Year, Month, Day, &SysTime );
		GetDateFormat( LOCALE_USER_DEFAULT, DATE_USE_ALT_CALENDAR, &SysTime, NULL, outString, SHORT_BUF_LEN );
		break;

	case 35 :

		shortbuf[0] = '\0';
		GetEnvironmentVariable( L"DateForm", shortbuf, SHORT_BUF_LEN );

		MakeSysTime( Year, Month, Day, &SysTime );

		if ( shortbuf[0] )
			GetDateFormat( LOCALE_USER_DEFAULT, 0, &SysTime, shortbuf, outString, SHORT_BUF_LEN );
		else
			GetDateFormat( LOCALE_USER_DEFAULT, DATE_LONGDATE, &SysTime, NULL, outString, SHORT_BUF_LEN );

		break;

	case 36 :

		MakeSysTime( Year, Month, Day, &SysTime );
		GetDateFormat( LOCALE_USER_DEFAULT, DATE_YEARMONTH, &SysTime, NULL, outString, SHORT_BUF_LEN );
		break;

	case 145 :

		GetLocalMonthAndDowNames();

		if ( LocalLongDateFmt() == 2 )
			wsprintf( outString, L"%s, %u %s %u", LocalDowNameLong[Julian( Year, Month, Day ) % 7], Day, LocalMonthNameLong[Month], Year );
		else
			wsprintf( outString, L"%s, %s %u, %u", LocalDowNameLong[Julian( Year, Month, Day ) % 7], LocalMonthNameLong[Month], Day, Year );

		break;

	case 4 :
	default :

		wsprintf( outString, L"%04u-%02u-%02u", Year, Month, Day );

	}

	return 0;
}


#endif	//	DATE_TO_STRING_VER


/*

	1.0.0.0		2022-12-17

*/
